/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>

#include "misc.h"
#include "regs.h"
#include "memory.h"
#include "loader.h"
#include "ss.h"
#include "endian.h"
#include "version.h"
#include "sim.h"

/* stats signal handler */
static void
signal_sim_stats(int sigtype)
{
  sim_dump_stats = TRUE;
}

/* exit signal handler */
static void
signal_exit_now(int sigtype)
{
  sim_exit_now = TRUE;
}

static void
usage(int argc, char **argv)
{
  fprintf(stderr, "Usage: %s [-options] exec args...\n", argv[0]);
  exit(1);
}

void
exit_now(int exit_code)
{
  sim_stats(stderr);
  mem_stats(stderr);
  exit(exit_code);
}

/* execution start time */
time_t start_time;

/* byte/word swapping required to execute target executable on this host */
int sim_swap_bytes;
int sim_swap_words;

void
main(int argc, char **argv, char **envp)
{
  int i, exit_code, exec_index;
  char c, *all_options, *s;
  extern char *strrchr(char *, char);

  /* opening banner */
  fprintf(stderr,
	  "%s: Version %d.%d.%d of %s.\n"
	  "Copyright (c) 1994-1996 by Todd M. Austin.  All Rights Reserved.\n",
	  ((s = strrchr(argv[0], '/')) ? s+1 : argv[0]),
	  VER_MAJOR, VER_MINOR, VER_REVISION, VER_UPDATE);

  /* print out the program arguments */
  fprintf(stderr, "**ARGS**: ");
  for (i=0; i<argc; i++)
    fprintf(stderr, "%s ", argv[i]);
  fprintf(stderr, "\n");

  if (argc < 2)
    usage(argc, argv);

  /* catch SIGUSR1 and dump intermediate stats */
  signal(SIGUSR1, signal_sim_stats);

  /* catch SIGUSR1 and dump intermediate stats */
  signal(SIGUSR2, signal_exit_now);

  /* register an error handler */
  fatal_hook(sim_stats);

  /* set up a non-local exit point */
  if ((exit_code = setjmp(sim_exit_buf)) != 0)
    {
      /* special handling as longjmp cannot pass 0 */
      exit_now(exit_code-1);
    }

  /* compute legal options */
  all_options = getopt_combine_options("v", sim_optstring);
  all_options = getopt_combine_options(all_options, mem_optstring);

  /* parse global options */
  getopt_init();
  while ((c = getopt_next(argc, argv, all_options)) != EOF)
    {
      switch (c) {
      case 'v':
	verbose = TRUE;
	break;
      case '?':
	usage(argc, argv);
      }
    }
  exec_index = getopt_index;

  /* parse simulator options */
  sim_options(argc, argv);
  mem_options(argc, argv);

  if (exec_index == argc)
    usage(argc, argv);

  /* initialize the instruction decoder */
  ss_init_decoder();

  /* initialize address spaces */
  for (i=0; i<MEM_TABLE_SIZE; i++)
    mem_table[i] = NULL;

  /* load the program text and data, set up environment, memory, and regs */
  ld_load_prog(mem_access, argc-exec_index,
	       argv+exec_index, envp, TRUE);

  /* initialize address spaces */
  regs_init();

  /* initialize processor precise state */
  regs_R[SS_STACK_REGNO] = ld_environ_base;
  regs_PC = ld_prog_entry;

  /* initialize address spaces */
  mem_init();

  /* record start of execution time, used in rate stats */
  start_time = time((time_t *)NULL);

  /* output simulation conditions */
  sim_config(stderr);
  mem_config(stderr);

  sim_main();

  panic("returned to main");
}
